-- Name of the private tooltip defined in TooltipIlvl.xml
TooltipIlvlTooltipName = "TooltipIlvlPrivateTooltip"

-- Set up caching for items
TooltipIlvlItemCache = nil

-- Set to true once initialization completes
local TooltipIlvlIsInitialized

-- Set up some constants
TOOLTIPILVL_CONSTANTS = {
	ADDON_NAME = "[Tooltip Ilvl]",
	FONTS      = {
		SIZE   = 11,
		COLORS = {
			TITLE   = "|cff0070dd",
			NORMAL  = "|cffffffff",
			TOOLTIP = "|cffffd700"
		}
	}
}

local TooltipIlvlFrame = CreateFrame("GameTooltip", TooltipIlvlTooltipName, WorldFrame, "GameTooltipTemplate")
TooltipIlvlFrame:SetOwner(WorldFrame, "ANCHOR_NONE")
TooltipIlvlFrame:ClearLines()
TooltipIlvlFrame:SetText("A")
TooltipIlvlFrame:Show()
TooltipIlvlFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
TooltipIlvlFrame:SetScript(
	"OnEvent",
	function(Frame, Event)
		Frame:UnregisterEvent(Event)
		Frame:SetScript("OnEvent", nil)
		TooltipIlvlOnEvent(Event)
	end
)

function TooltipIlvlOnEvent(Event)
	if Event == "PLAYER_ENTERING_WORLD" then
		TooltipIlvlInitialize()
	end
end

-- Initializes Tooltip Ilvl after saved variables have been loaded
function TooltipIlvlInitialize()
	if TooltipIlvlIsInitialized then return end

	-- Hook up all of the events we'll need
	hooksecurefunc(GameTooltip, "SetHyperlink", function(self, ...) TooltipIlvlUpdateTooltip("GameTooltip", "SetHyperlink", ...) end)
	hooksecurefunc(GameTooltip, "SetBagItem", function(self, ...) TooltipIlvlUpdateTooltip("GameTooltip", "SetBagItem", ...) end)
	hooksecurefunc(GameTooltip, "SetInboxItem", function(self, ...) TooltipIlvlUpdateTooltip("GameTooltip", "SetInboxItem", ...) end)
	hooksecurefunc(GameTooltip, "SetInventoryItem", function(self, ...) TooltipIlvlUpdateTooltip("GameTooltip", "SetInventoryItem", ...) end)
	hooksecurefunc(GameTooltip, "SetItemByID", function(self, ...) TooltipIlvlUpdateTooltip("GameTooltip", "SetItemByID", ...) end)
	hooksecurefunc(GameTooltip, "SetLootItem", function(self, ...) TooltipIlvlUpdateTooltip("GameTooltip", "SetLootItem", ...) end)
	hooksecurefunc(GameTooltip, "SetLootRollItem", function(self, ...) TooltipIlvlUpdateTooltip("GameTooltip", "SetLootRollItem", ...) end)
	hooksecurefunc(GameTooltip, "SetMerchantItem", function(self, ...) TooltipIlvlUpdateTooltip("GameTooltip", "SetMerchantItem", ...) end)
	hooksecurefunc(GameTooltip, "SetQuestItem", function(self, ...) TooltipIlvlUpdateTooltip("GameTooltip", "SetQuestItem", ...) end)
	hooksecurefunc(GameTooltip, "SetQuestLogItem", function(self, ...) TooltipIlvlUpdateTooltip("GameTooltip", "SetQuestLogItem", ...) end)
	hooksecurefunc(GameTooltip, "SetSendMailItem", function(self, ...) TooltipIlvlUpdateTooltip("GameTooltip", "SetSendMailItem", ...) end)
	hooksecurefunc(GameTooltip, "SetBuybackItem", function(self, ...) TooltipIlvlUpdateTooltip("GameTooltip", "SetBuybackItem", ...) end)

	-- Hook into events for item comparisons, setting up 2 for instances like rings and trinkets
	hooksecurefunc(ShoppingTooltip1, "SetCompareItem",
		function(self, ...)
			local _, ItemLink1 = ShoppingTooltip1:GetItem()
			TooltipIlvlUpdateTooltip("ShoppingTooltip1", "SetCompareItem", ItemLink1, ...)
			local _, ItemLink2 = ShoppingTooltip2:GetItem()
			if ItemLink2 and ShoppingTooltip2:IsShown() then
				TooltipIlvlUpdateTooltip("ShoppingTooltip2", "SetHyperlink", ItemLink2, ...)
			end
		end)
	hooksecurefunc(ItemRefShoppingTooltip1, "SetCompareItem",
		function(self, ...)
			local _, ItemLink1 = ItemRefShoppingTooltip1:GetItem()
			TooltipIlvlUpdateTooltip("ItemRefShoppingTooltip1", "SetCompareItem", ItemLink1, ...)
			local _, ItemLink2 = ItemRefShoppingTooltip2:GetItem()
			if ItemLink2 and ItemRefShoppingTooltip2:IsShown() then
				TooltipIlvlUpdateTooltip("ItemRefShoppingTooltip2", "SetHyperlink", ItemLink2, ...)
			end
		end)

	-- If we don't already have this saved to a cache file then create it
	if not TooltipIlvlOptions then TooltipIlvlOptions = {} end

	-- Update to true so we don't load everything again
	TooltipIlvlIsInitialized = true
	TooltipIlvlSendMessage("Addon loaded!")
end

function TooltipIlvlUpdateTooltip(TooltipName, MethodName, Param1, ...)
	local Tooltip = _G[TooltipName]
	if not Tooltip then
		TooltipIlvlSendMessage("The tooltip is missing, which shouldn't happen")
		return
	end

	-- Start by getting the item link.
	local ItemLink = TooltipIlvlGetItemLinkFromTooltip(TooltipName, MethodName, Param1, ...)

	local Item

	if ItemLink then
		Item = { TooltipIlvlGetItemData(ItemLink) }
	else
		-- If there was no item link, try reading directly from the tooltip.
		Item = { TooltipIlvlGetItemDataFromTooltip(TooltipName, MethodName, Param1, ...) }
	end

	if Item then
		local ItemLevel, ItemType = Item[4], _G[Item[9]]

		if ItemLevel and ItemType then
			for i = 1, Tooltip:NumLines() do
				if _G[Tooltip:GetName().."TextLeft"..i]:GetText() == ItemType then
					TooltipIlvlInsertTooltipLine(Tooltip, format("Item Level %s", ItemLevel))
					break
				end
			end
		end
	end
end

function TooltipIlvlGetItemLinkFromTooltip(TooltipName, MethodName, Param1, ...)
	if (not TooltipName) or (not MethodName) then return end

	-- First, find the tooltip.
	local Tooltip = _G[TooltipName]
	if not Tooltip then return end

	-- If we have a tooltip, try to get an item link from it.
	local ItemLink, _
	if (MethodName == "SetHyperlink") and Param1 then
		-- Special case: if the method is SetHyperlink, then we already have an item link.
		-- (Normally, GetItem will work, but SetHyperlink is used by some mod compatibility code.)
		ItemLink = Param1
	elseif Tooltip.GetItem then
		_, ItemLink = Tooltip:GetItem()
	end

	return ItemLink
end

-- Gets the item data for a specific item link. Retrieves the information from the cache when possible; otherwise, it gets fresh information.
function TooltipIlvlGetItemData(ItemLink)
	if not ItemLink then return end

	-- Only item links are supported; other links are not.
	if TooltipIlvlGetHyperlinkType(ItemLink) ~= "item" then return end

	-- If this type of item isn't equippable then do nothing
	if not TooltipIlvlItemIsEquipable(ItemLink) then return end

	local ItemId = tonumber(ItemLink:match("item:(%d+):"))

	-- If we have an item link, we can extract basic data from it from the user's WoW cache (not the Tooltip Ilvl item cache).
	-- We get a new, normalized version of ItemLink so that items don't end up in the cache multiple times if they're requested
	-- using different styles of links that all point to the same item.
	local _, NewItemLink, _, _ = GetItemInfo(ItemLink)
	if NewItemLink then
		ItemLink = NewItemLink
	end

	-- Look up the item in the TooltipIlvl item cache.
	local CachedItem = TooltipIlvlGetCachedItem(ItemId)
	if not CachedItem and not NewItemLink then
		-- The item isn't in the user's WoW cache or TooltipIlvl cache, so do nothing.
		return
	end
	if CachedItem and CachedItem.Values then
		return CachedItem
	end

	-- Cache this item so we don't have to re-parse next time.
	TooltipIlvlCacheItem(ItemId)

	return GetItemInfo(ItemLink)
end

-- Gets the item data for a specific item.  Retrieves the information from the cache when possible; otherwise, gets it from the tooltip specified.  Only use this if you don't have an item link; otherwise use TooltipIlvlGetItemData(ItemLink).
-- Return value type is the same as TooltipIlvlGetCachedItem.
function TooltipIlvlGetItemDataFromTooltip(TooltipName, MethodName, Param1, ...)
	local ItemName, ItemNameLineNumber = TooltipIlvlGetItemNameFromTooltip(TooltipName)
	if (not ItemName) or (not ItemNameLineNumber) then return end
	local ItemLink = TooltipIlvlGetItemLinkFromTooltip(TooltipName, MethodName, Param1, ...)
	print(ItemLink)
	local Item = TooltipIlvlGetCachedItem(ItemLink)
	if Item and Item.Values then
		return Item
	end

	TooltipIlvlCacheItem(ItemLink)

	return GetItemInfo(ItemLink)
end

-- Gets the name of an item given a tooltip name, and the line on which the item appears.
-- Normally this is line 1, but it can be line 2 if the first line is "Currently Equipped".
-- Parameters: TooltipName
--		TooltipName: The name of the tooltip to read.
-- Return value: ItemName, LineNumber
--		ItemName: The name of the item in the tooltip, or nil if the tooltip didn't have one.
--		LineNumber: The line number on which the name was found, or nil if no item was found.
function TooltipIlvlGetItemNameFromTooltip(TooltipName)
	-- First, get the tooltip details.
	local TooltipTopLine = _G[TooltipName .. "TextLeft1"]
	if not TooltipTopLine then return end
	local ItemName = TooltipTopLine:GetText()
	if not ItemName or ItemName == "" or ItemName == RETRIEVING_ITEM_INFO then return end
	
	-- If this is a Currently Equipped tooltip, skip the first line.
	if ItemName == CURRENTLY_EQUIPPED then
		ItemNameLineNumber = 2
		TooltipTopLine = _G[TooltipName .. "TextLeft2"]
		if not TooltipTopLine then return end
		return TooltipTopLine:GetText(), 2
	end
	return ItemName, 1
end

-- Searches the item cache for an item, and either returns the correct cached item, or nil.
function TooltipIlvlGetCachedItem(ItemId)
	local _
	if TooltipIlvlOptions.ItemCache[ItemId] then
		return unpack(TooltipIlvlOptions.ItemCache[ItemId])
	end

	return
end

-- Adds an item to the item cache
function TooltipIlvlCacheItem(ItemId)
	if not TooltipIlvlOptions.ItemCache then TooltipIlvlOptions.ItemCache = {} end
	TooltipIlvlOptions.ItemCache[ItemId] = {GetItemInfo(ItemId)}
end

-- Returns the type of hyperlink passed in, or nil if it's not a hyperlink.
-- Possible values include: item, enchant, quest, spell
function TooltipIlvlGetHyperlinkType(Hyperlink)
	-- First, try colored links.
	local _, _, LinkType = strfind(Hyperlink, "^|c%x%x%x%x%x%x%x%x|H(.-):")
	if not LinkType then
		-- Then, try links prepended with |H.  (Outfitter does this.)
		_, _, LinkType = strfind(Hyperlink, "^|H(.-):")
	end
	if not LinkType then
		-- Then, try raw links.
		_, _, LinkType = strfind(Hyperlink, "^(.-):")
	end
	return LinkType
end

-- Insert a new line to the toolip with the item level
function TooltipIlvlInsertTooltipLine(Tooltip, Line)
	-- Add a blank line and then the actual text so we're not clashing with other tooltips
	Tooltip:AddLine(" ")
	Tooltip:AddLine(Line, 1, 0.84, 0)
	Tooltip:Show()
end

-- Determine whether or not an item is equippable
function TooltipIlvlItemIsEquipable(ItemLink)
	local _, _, _, InvType, _, _, _ = GetItemInfoInstant(ItemLink)
	return InvType ~= nil and InvType ~= ""
end

-- Send a message to the users chat
function TooltipIlvlSendMessage(Message)
    print(TOOLTIPILVL_CONSTANTS.FONTS.COLORS.TITLE .. TOOLTIPILVL_CONSTANTS.ADDON_NAME .. " " .. TOOLTIPILVL_CONSTANTS.FONTS.COLORS.NORMAL .. Message)
end